module Queue(Qu, empty, enq, first, deq) where
import Data.List(intercalate)

-- A Queue is given by two lists
data Qu alpha = Qu [alpha] [alpha]

-- The empty Q
empty :: Qu alpha
empty = Qu [] []

-- Enqueue 
enq   :: alpha-> Qu alpha-> Qu alpha
enq x (Qu xs ys) = check xs (x:ys)  

-- Front of Q
first :: Qu alpha-> alpha  
first (Qu [] _)     = error "Queue: first of empty Q"
first (Qu (x:xs) _) = x

-- Dequeue
deq   :: Qu alpha-> Qu alpha
deq (Qu [] _ )     = error "Queue: deq of empty Q"
deq (Qu (_:xs) ys) = check xs ys

-- Check and restore invariant
check :: [alpha]-> [alpha]-> Qu alpha
check [] ys = Qu (reverse ys) []
check xs ys = Qu xs ys

-- A queue is valid if the invariant holds
valid :: Qu alpha-> Bool
valid (Qu [] ys)  = null ys
valid (Qu (_:_) _) = True

-- Equality for queues.
-- Two queues are equal iff both are valid, and their content is the same.
instance Eq alpha=> Eq (Qu alpha) where
  q1@(Qu xs1 ys1) == q2@(Qu xs2 ys2) =
    valid q1 && valid q2 &&
    xs1 ++ reverse ys1 == xs2 ++ reverse ys2

-- Show (needed with quickcheck to show failures)
-- Displays front of queue to the right, as on the slides
instance Show a=> Show (Qu a) where
  show (Qu xs ys) =
    "<"++ intercalate "->" (map show ys)
     ++ "->" ++ intercalate "->" (map show $ reverse xs) ++ ">"
